@@ -11,6 +11,9 @@ from typing import (
     cast,
     Dict,
     Any,
+    Generic,
+    TypeVar,
+    Iterator,
 )
 from collections import defaultdict
 from operator import itemgetter
@@ -20,14 +23,17 @@ from .exceptions import MultipleRootError, NotFoundNodeError, DuplicatedNodeErro
 from .utils import STYLES
 
 
-# root has no key (None), keyed node has children with str keys, unkeyed node has children with int keys
-Key = Union[None, str, int]
-KeyedNode = Tuple[Key, Node]
-KeyedTree = Tuple[Key, "Tree"]
-Path = str
+# keyed node has children with str keys, unkeyed node has children with int keys
+# note: root has no key (None value)
+Key = Union[str, int]
+KeyedTree = Tuple[Optional[Key], "Tree"]
+Path = Iterable[Key]
+
+GenericNode = TypeVar("GenericNode", bound=Node)
+KeyedNode = Tuple[Optional[Key], GenericNode]
 
 
-class Tree(object):
+class Tree(Generic[GenericNode]):
 
     """Principles:
     - each node is identified by an id
@@ -43,13 +49,11 @@ class Tree(object):
 
     """
 
-    def __init__(self, path_separator: str = ".") -> None:
-        self.path_separator = path_separator
-
+    def __init__(self) -> None:
         # nodes references and hierarchy in tree
         self.root: Optional[NodeId] = None
         # node identifier -> node
-        self._nodes_map: Dict[NodeId, Node] = {}
+        self._nodes_map: Dict[NodeId, GenericNode] = {}
         # node identifier -> parent node identifier
         self._nodes_parent: Dict[NodeId, Optional[NodeId]] = defaultdict(lambda: None)
         # "map" node identifier -> map of children nodes identifier -> key
@@ -61,10 +65,6 @@ class Tree(object):
         return identifier in self._nodes_map
 
     def get(self, nid: NodeId) -> KeyedNode:
-        """Get a node by its id.
-        :param nid: str, identifier of node to fetch
-        :rtype: lighttree.node.Node
-        """
         self._ensure_present(nid)
         return self.get_key(nid), self._nodes_map[nid]
 
@@ -78,37 +78,40 @@ class Tree(object):
             if child_id is None:
                 raise ValueError("No child of key %s below %s" % (key, nid))
             return child_id
-        if not isinstance(key, (str, int)):
-            raise ValueError("Expected integer 'castable' key, got %s" % key)
-        return self._nodes_children_list[nid][int(key)]
+        try:
+            return self._nodes_children_list[nid][int(key)]
+        except (KeyError, ValueError, TypeError):
+            raise ValueError("No child of key %s below %s" % (key, nid))
 
     def child(self, nid: NodeId, key: Key) -> KeyedNode:
         return self.get(self.child_id(nid, key))
 
-    def get_node_id_by_path(self, path: Path) -> NodeId:
+    def get_node_id_by_path(self, path: Path, strict: bool = True) -> NodeId:
         nid = self.root
         if nid is None:
             raise ValueError("Empty tree")
         if path == "":
             return nid
-        keys = str(path).split(self.path_separator)
-        for k in keys:
-            nid = self.child_id(nid, k)
+        for k in path:
+            try:
+                nid = self.child_id(nid, k)
+            except ValueError:
+                if strict or not isinstance(k, str) or not k.isdigit():
+                    raise
+                nid = self.child_id(nid, int(k))
         if nid is None:
-            raise ValueError("Empty tree")
+            raise ValueError("Not found")
         return nid
 
     def get_path(self, nid: NodeId) -> Path:
-        return self.path_separator.join(
-            [
-                str(k)
-                for k, _ in self.ancestors(nid, from_root=True, include_current=True)[
-                    1:
-                ]
-            ]
-        )
+        return [
+            # ignore typing warning of potential None value, since None only applies at root node which is excluded
+            # [1:] -> exclude root node key
+            k  # type: ignore
+            for k, _ in self.ancestors(nid, from_root=True, include_current=True)[1:]
+        ]
 
-    def get_key(self, nid: NodeId) -> Key:
+    def get_key(self, nid: NodeId) -> Optional[Key]:
         """Get a node's key.
         :param nid: str, identifier of node
 
@@ -128,13 +131,12 @@ class Tree(object):
         self,
         id_in: Optional[Sequence[NodeId]] = None,
         depth_in: Optional[Sequence[int]] = None,
-        filter_: Optional[Callable[[Node], bool]] = None,
+        filter_: Optional[Callable[[GenericNode], bool]] = None,
     ) -> List[KeyedNode]:
         """List nodes.
         :param id_in: list of str, optional, filter nodes among provided identifiers
         :param depth_in: list of int, optional, filter nodes whose depth in tree is among provided values
         :param filter\_: function, optional, filtering function to apply to each node
-        :rtype: list of lighttree.node.Node
         """
         return [
             (self.get_key(nid), node)
@@ -167,7 +169,7 @@ class Tree(object):
             raise NotFoundNodeError("Node id <%s> doesn't exist in tree" % nid)
         return nid
 
-    def _validate_node_insertion(self, node: Node) -> None:
+    def _validate_node_insertion(self, node: GenericNode) -> None:
         if node.identifier in self._nodes_map.keys():
             raise DuplicatedNodeError(
                 "Can't create node with id '%s'" % node.identifier
@@ -330,24 +332,12 @@ class Tree(object):
 
     def insert(
         self,
-        item: Union[Node, "Tree"],
+        item: Union[GenericNode, "Tree"],
         parent_id: Optional[NodeId] = None,
         child_id: Optional[NodeId] = None,
         child_id_below: Optional[NodeId] = None,
-        key: Key = None,
+        key: Optional[Key] = None,
     ) -> "Tree":
-        if isinstance(item, Node):
-            if child_id_below is not None:
-                raise ValueError(
-                    '"child_id_below" parameter is reserved to Tree insertion.'
-                )
-            self.insert_node(
-                node=item,
-                parent_id=parent_id,
-                child_id=child_id,
-                key=key,
-            )
-            return self
         if isinstance(item, Tree):
             self.insert_tree(
                 new_tree=item,
@@ -357,24 +347,26 @@ class Tree(object):
                 key=key,
             )
             return self
-        raise ValueError(
-            '"item" parameter must either be a Node, or a Tree, got <%s>.' % type(item)
+        # item is GenericNode
+        if child_id_below is not None:
+            raise ValueError(
+                '"child_id_below" parameter is reserved to Tree insertion.'
+            )
+        self.insert_node(
+            node=item,
+            parent_id=parent_id,
+            child_id=child_id,
+            key=key,
         )
+        return self
 
     def insert_node(
         self,
-        node: Node,
+        node: GenericNode,
         parent_id: Optional[NodeId] = None,
         child_id: Optional[NodeId] = None,
-        key: Key = None,
-    ) -> Key:
-        """Insert node, return key
-        :param node:
-        :param parent_id:
-        :param child_id:
-        :param key:
-        :return:
-        """
+        key: Optional[Key] = None,
+    ) -> Optional[Key]:
         self._validate_node_insertion(node)
         if parent_id is not None and child_id is not None:
             raise ValueError('Can declare at most "parent_id" or "child_id"')
@@ -386,9 +378,9 @@ class Tree(object):
 
     def _insert_node_below(
         self,
-        node: Node,
+        node: GenericNode,
         parent_id: Optional[NodeId],
-        key: Key,
+        key: Optional[Key],
     ) -> None:
         # insertion at root
         if parent_id is None:
@@ -433,7 +425,9 @@ class Tree(object):
         self._nodes_map[node_id] = node
         self._nodes_parent[node_id] = parent_id
 
-    def _insert_node_above(self, node: Node, child_id: NodeId, key: Key) -> None:
+    def _insert_node_above(
+        self, node: GenericNode, child_id: NodeId, key: Optional[Key]
+    ) -> None:
         self._ensure_present(child_id)
         # get parent_id before dropping subtree
         try:
@@ -457,9 +451,8 @@ class Tree(object):
         parent_id: Optional[NodeId] = None,
         child_id: Optional[NodeId] = None,
         child_id_below: Optional[NodeId] = None,
-        key: Key = None,
-    ) -> Key:
-        """Return new key"""
+        key: Optional[Key] = None,
+    ) -> Optional[Key]:
         self._validate_tree_insertion(new_tree)
         if new_tree.root is None:
             raise ValueError("Empty inserted tree")
@@ -483,7 +476,7 @@ class Tree(object):
         self,
         new_tree: "Tree",
         parent_id: Optional[NodeId],
-        key: Key,
+        key: Optional[Key],
     ) -> "Tree":
         if parent_id is None:
             # insertion at root requires tree to be empty
@@ -509,7 +502,7 @@ class Tree(object):
         new_tree: "Tree",
         child_id: NodeId,
         child_id_below: Optional[NodeId],
-        key: Key,
+        key: Optional[Key],
     ) -> None:
         # make all checks before modifying tree
         self._ensure_present(child_id)
@@ -601,10 +594,10 @@ class Tree(object):
         self,
         nid: Optional[NodeId] = None,
         mode: str = "depth",
-        filter_: Optional[Callable[[Union[None, str, int], Node], bool]] = None,
+        filter_: Optional[Callable[[Union[None, str, int], GenericNode], bool]] = None,
         filter_through: bool = False,
         reverse: bool = False,
-    ) -> Iterable[KeyedNode]:
+    ) -> Iterator[KeyedNode]:
         """Python generator traversing the tree (or a subtree) with optional node filtering.
 
         Inspired by treelib implementation https://github.com/caesar0301/treelib/blob/master/treelib/tree.py#L374
@@ -615,7 +608,6 @@ class Tree(object):
         :param filter_through: if True, excluded nodes don't exclude their children.
         :param reverse: the ``reverse`` param for sorting :class:`Node` objects in the same level
         :return: node ids that satisfy the conditions if ``id_only`` is True, else nodes.
-        :rtype: generator
         """
         if mode not in ("depth", "width"):
             raise NotImplementedError("Traversal mode '%s' is not supported" % mode)
@@ -656,7 +648,7 @@ class Tree(object):
     def show(
         self,
         nid: Optional[NodeId] = None,
-        filter_: Optional[Callable[[Node], bool]] = None,
+        filter_: Optional[Callable[[GenericNode], bool]] = None,
         display_key: bool = True,
         reverse: bool = False,
         line_type: str = "ascii-ex",
@@ -676,8 +668,6 @@ class Tree(object):
         :param limit: int, truncate tree display to this number of lines
         :param kwargs: kwargs params passed to node ``line_repr`` method
         :param line_max_length
-        :rtype: unicode in python2, str in python3
-
         """
         output = ""
 
@@ -714,10 +704,10 @@ class Tree(object):
     def _iter_nodes_with_location(
         self,
         nid: Optional[NodeId],
-        filter_: Optional[Callable[[Node], bool]],
+        filter_: Optional[Callable[[GenericNode], bool]],
         reverse: bool,
         is_last_list: Optional[List[bool]] = None,
-    ) -> Iterable[Tuple[Tuple[bool, ...], Key, Node]]:
+    ) -> Iterable[Tuple[Tuple[bool, ...], Optional[Key], GenericNode]]:
         """Yield nodes with information on how they are placed.
         :param nid: starting node identifier
         :param filter_: filter function applied on nodes
